﻿using ISONE.WindIntegration.WebServices.WSPlantService;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Net.Security;
using System.Security.Cryptography.X509Certificates;
using System.ServiceModel;
using System.Text;
using System.Threading.Tasks;

namespace WindIntegrationSampleClient
{
    class Program
    {
        static void Main(string[] args)
        {
            try
            {
                // Setup the proxy to the web service.
                WindPlantLeadParticipantPortType proxy = CreateWebServiceClient();
                // Try a simple Query and Submit type operation.
                TestQueryCategories(proxy);
                TestQueryEntities(proxy);
                TestQuerySchedules(proxy);
                TestQueryForecast(proxy);
                TestSubmitForecast(proxy);

                proxy = null;
            }

            catch (FaultException<ISONE.WindIntegration.WebServices.WSPlantService.FaultType> soapFault)
            {
                Console.WriteLine("\n*** Operation produced a Fault ***\n");
                // display details of soap message
                foreach (ISONE.WindIntegration.WebServices.WSPlantService.TransactionVariance tv in soapFault.Detail)
                    Console.WriteLine("ErrorMessage: {0}", tv.message);
            }
            catch (Exception e)
            {
                Console.WriteLine("\n*** Operation produced a Fault ***\n");
                Console.WriteLine(e.Message);
            }

            //Uncomment to have program exit automatically
            //Environment.Exit(0);

            //Uncomment to have program stay open until 'Enter' key is pressed.
            Console.WriteLine("\nPress 'Enter' to exit..\n");
            Console.ReadLine();
        }

        //---------------------------------------------------------------------
        ///<summary>
        /// Executes a "TestQueryCategories" operation and prints out the results.
        ///</summary>
        //---------------------------------------------------------------------
        private static void TestQueryCategories(WindPlantLeadParticipantPortType proxy)
        {
            Console.WriteLine("\n-----------------------");
            Console.WriteLine("  Testing QueryCategories");
            Console.WriteLine("\n-----------------------\n");

            // Call the service and print out the response.
            QueryCategoriesRequest request = new QueryCategoriesRequest();
            QueryCategoriesResponse response = proxy.QueryCategories(request);
            int itemCount = 0;
            if (response != null && response.Categories != null && response.Categories.Length > 0)
            { itemCount = response.Categories.Length; }

            Console.Write("returned: {0}. ", itemCount);
            Console.WriteLine("Here are the first 10:\n");
            int loopCount = 0;
            foreach (CategoryIdentityType obj in response.Categories)
            {
                Console.WriteLine("ID = {0,-8} Name = {1,-25} Desc = {2}",
                                   obj.identifier, obj.name, obj.description);
                loopCount++;
                if (loopCount == 10)
                    break;
            }
        }

        //---------------------------------------------------------------------
        ///<summary>
        /// Executes a "TestQueryEntities" operation and prints out the results.
        ///</summary>
        //---------------------------------------------------------------------
        private static void TestQueryEntities(WindPlantLeadParticipantPortType proxy)
        {
            Console.WriteLine("\n-----------------------");
            Console.WriteLine("  Testing QueryEntities");
            Console.WriteLine("\n-----------------------\n");

            // Call the service and print out the response.
            QueryEntitiesRequest request = new QueryEntitiesRequest();
            request.Category = new CategoryIdentityType();
            request.Category.identifier = "WPLANT";
            QueryEntitiesResponse response = proxy.QueryEntities(request);
            int itemCount = 0;
            if (response != null && response.Entities != null && response.Entities.Length > 0)
            { itemCount = response.Entities.Length; }

            Console.Write("returned: {0}. ", itemCount);
            Console.WriteLine("Here are the first 10:\n");
            int loopCount = 0;
            foreach (EntityIdentityType obj in response.Entities)
            {
                Console.WriteLine("ID = {0,-8} Name = {1,-25} Desc = {2}",
                                   obj.identifier, obj.name, obj.description);
                loopCount++;
                if (loopCount == 10)
                    break;
            }
        }

        //---------------------------------------------------------------------
        ///<summary>
        /// Executes a "TestQuerySchedules" operation and prints out the results.
        ///</summary>
        //---------------------------------------------------------------------
        private static void TestQuerySchedules(WindPlantLeadParticipantPortType proxy)
        {
            Console.WriteLine("\n-----------------------");
            Console.WriteLine("  Testing QuerySchedules");
            Console.WriteLine("\n-----------------------\n");

            // Call the service and print out the response.
            QuerySchedulesRequest request = new QuerySchedulesRequest();
            QuerySchedulesResponse response = proxy.QuerySchedules(request);
            int itemCount = 0;
            if (response != null && response.Schedules != null && response.Schedules.Length > 0)
            { itemCount = response.Schedules.Length; }

            Console.Write("returned: {0}. ", itemCount);
            Console.WriteLine("Here are the first 10:\n");
            int loopCount = 0;
            foreach (ScheduleIdentityType obj in response.Schedules)
            {
                Console.WriteLine("ID = {0,-8} Name = {1,-25} Desc = {2}",
                                   obj.identifier, obj.name, obj.description);
                loopCount++;
                if (loopCount == 10)
                    break;
            }
        }

        //---------------------------------------------------------------------
        ///<summary>
        /// Executes a "TestQueryForecast" operation and prints out the results.
        ///</summary>
        //---------------------------------------------------------------------
        private static void TestQueryForecast(WindPlantLeadParticipantPortType proxy)
        {
            Console.WriteLine("\n-----------------------");
            Console.WriteLine("  Testing QueryForecast");
            Console.WriteLine("\n-----------------------\n");

            // Call the service and print out the response.
            DateRangeType dateRangeType = new DateRangeType();
            dateRangeType.fromTime = DateTime.Now.AddMinutes(-60);
            dateRangeType.toTime = DateTime.Now.AddMinutes(60);

            EntityIdentityType entityIdentityType = new EntityIdentityType();
            // ***** NOTE Could use Entity Identifier OR Asset Identifier **********
            entityIdentityType.identifier = 11111; // TO DO Use the correct Entity Identifier
            entityIdentityType.identifierSpecified = true;

            //entityIdentityType.assetIdentifier = 22222; // TO DO Use the correct Asset Identifier
            //entityIdentityType.assetIdentifierSpecified = true;

            EntityIdentityType[] entityIdentityTypeArray = new EntityIdentityType[1];
            entityIdentityTypeArray[0] = entityIdentityType;

            ScheduleIdentityType scheduleIdentityType = new ScheduleIdentityType();
            scheduleIdentityType.identifier = 1234567890123; // TO DO Use the correct Schedule Identifier

            BaseRequestType baseRequestType = new BaseRequestType();
            baseRequestType.Schedule = scheduleIdentityType;
            baseRequestType.Entities = entityIdentityTypeArray;
            baseRequestType.TimeRange = dateRangeType;

            BaseRequestType[] baseRequestTypeArray = new BaseRequestType[1];
            baseRequestTypeArray[0] = baseRequestType;

            QueryForecastRequest request = new QueryForecastRequest(baseRequestTypeArray);

            QueryForecastResponse response = proxy.QueryForecast(request);
            int itemCount = 0;
            if (response != null && response.ScheduleResponse != null && response.ScheduleResponse.Length > 0)
            { itemCount = response.ScheduleResponse.Length; }

            Console.Write("returned: {0}. ", itemCount);
            Console.WriteLine("Here are the first 10 Time/Value Series:\n");
            int loopCount = 0;
            foreach (PowerScheduleDataType obj in response.ScheduleResponse)
            {
                foreach (ScheduleIdentityType sch in obj.Schedule)
                    Console.WriteLine("Schedule ID = {0} Name = {1}", sch.identifier, sch.name);
                foreach (PowerEntityIdentityType ent in obj.Entities)
                {
                    Console.WriteLine("Entity ID = {0} Name = {1} Asset{2}", ent.identifier, ent.name, ent.assetIdentifier);
                    foreach (TimeValueSeriesType tvs in ent.Power)
                    {
                        Console.WriteLine("time = {0} Value = {1}", tvs.time, tvs.value);
                        loopCount++;
                        if (loopCount == 10)
                            break;
                    }
                }
            }
        }

        //---------------------------------------------------------------------
        ///<summary>
        /// Executes a "TestSubmitForecast" operation and prints out the results.
        ///</summary>
        //---------------------------------------------------------------------
        const int resolution = 60;
        const int periodicity = 60;
        const int basetime = 0;
        const int fromforecasthorizon = 60;
        const int toforecasthorizon = 2880;
        const int mvarating = 22;
        private static void TestSubmitForecast(WindPlantLeadParticipantPortType proxy)
        {
            Console.WriteLine("\n-----------------------");
            Console.WriteLine("  Testing SubmitForecast");
            Console.WriteLine("\n-----------------------\n");

            TimeValueSeriesType[] timeValueSeriesTypeArray = dataGenerator(resolution, periodicity, basetime, fromforecasthorizon, toforecasthorizon);

            PowerEntityIdentityType powerEntityIdentityType = new PowerEntityIdentityType();
            // ***** NOTE Could use Entity Identifier OR Asset Identifier **********
            powerEntityIdentityType.identifier = 11111; // TO DO Use the correct Entity Identifier
            powerEntityIdentityType.identifierSpecified = true;

            //powerEntityIdentityType.assetIdentifier = 22222; // TO DO Use the correct Asset Identifier
            //powerEntityIdentityType.assetIdentifierSpecified = true;

            powerEntityIdentityType.Power = timeValueSeriesTypeArray;

            PowerEntityIdentityType[] powerEntityIdentityTypeArray = new PowerEntityIdentityType[1];
            powerEntityIdentityTypeArray[0] = powerEntityIdentityType;

            DateRangeType dateRangeType = new DateRangeType();
            dateRangeType.fromTime = DateTime.Now.AddMinutes(fromforecasthorizon);
            dateRangeType.toTime = DateTime.Now.AddMinutes(toforecasthorizon);

            ScheduleIdentityType scheduleIdentityType = new ScheduleIdentityType();
            scheduleIdentityType.identifier = 1234567890123; // TO DO Use the correct Schedule Identifier to submit WPFA

            ScheduleIdentityType[] scheduleIdentityTypeArray = new ScheduleIdentityType[1];
            scheduleIdentityTypeArray[0] = scheduleIdentityType;

            PowerScheduleDataType powerScheduleDataType = new PowerScheduleDataType();
            powerScheduleDataType.Schedule = scheduleIdentityTypeArray;
            powerScheduleDataType.TimeRange = dateRangeType;
            powerScheduleDataType.Entities = powerEntityIdentityTypeArray;

            PowerScheduleDataType[] powerScheduleDataTypeArray = new PowerScheduleDataType[1];
            powerScheduleDataTypeArray[0] = powerScheduleDataType;

            SubmitForecastRequest request = new SubmitForecastRequest();
            request.CreateSchedule = powerScheduleDataTypeArray;

            SubmitForecastResponse response = proxy.SubmitForecast(request);

            if(response !=null && response.SubmitStatusResponse != null && response.SubmitStatusResponse.success != null)
            {
                Console.WriteLine("Forecast Submitted successfully! Transaction ID = {0} ", response.SubmitStatusResponse.success.transactionId);
            }
        }

        public static TimeValueSeriesType[] dataGenerator(int resolution, int periodicity, int basetime, int fromforecasthorizon, int toforecasthorizon)
        {
            TimeValueSeriesType[] timeValueSeriesTypeArray = new TimeValueSeriesType[((toforecasthorizon - fromforecasthorizon) / resolution) + 1];

            DateTime now = DateTime.Now.ToUniversalTime();

            //Round to next interval
            now = now.AddMinutes(periodicity - ((now.Minute + now.Hour * 60) % periodicity)).AddMinutes(basetime).AddSeconds(-now.Second);

            //Fill in TimeValueSeries
            for (int i = 0; now.AddMinutes((i * resolution) + fromforecasthorizon) <= now.AddMinutes(toforecasthorizon); i++)
            {
                TimeValueSeriesType timeValueSeriesType = new TimeValueSeriesType();
                timeValueSeriesType.time = now.AddMinutes((i * resolution) + fromforecasthorizon);
                timeValueSeriesType.value = i % mvarating;

                timeValueSeriesTypeArray[i] = timeValueSeriesType;
            }

            return timeValueSeriesTypeArray;
        }

        //---------------------------------------------------------------------
        ///<summary>
        /// Setup the endpoints and certificates and return the proxy object
        /// for the web service.
        ///</summary>
        //---------------------------------------------------------------------
        private static WindPlantLeadParticipantPortType CreateWebServiceClient()
        {
            EndpointAddress address = null;
            try { address = new EndpointAddress("https://hostname/path");  /* //TO DO Update the correct End Point URL*/ }
            catch (Exception ex)
            {
                throw new Exception("Unable to create EndpointAddress.  \n" +
                                    "Check endpoint address \n" +
                                    ex.Message);
            }

            // The client will use a 509 certificate. 
            X509Certificate2 clientKey = null;
            try
            {
                clientKey = new X509Certificate2("test.p12", "test"); /* //TO DO Update the client certficate and password */
            }
            catch (Exception ex)
            {
                throw new Exception("Unable to create ClientCertificate. \n" +
                                     "Check valid certificate credentials \n" +
                                     ex.Message);
            }

            // And the server must have trusted certificates.
            X509Certificate2 serverCertifications = null;
            try
            { serverCertifications = new X509Certificate2("ca.crt"); /* //TO DO Update the server trust store cert */}
            catch (Exception ex)
            {
                throw new Exception("Unable to create TrustedServerCertificate. \n" +
                                       "Check server trust store cert \n" +
                                       ex.Message);
            }


            //-----------------------------------------------------------------
            // Establish a proxy for the service
            //-----------------------------------------------------------------

            WindPlantLeadParticipantPortType proxy = null;
            try
            {
                proxy = GetService(
                        address,
                        clientKey,
                        serverCertifications);
            }
            catch (Exception)
            {
                Console.WriteLine("Unable to create WindPlantLeadParticipantPortTypeClient proxy");
                throw;
            }

            return proxy;
        }

        //---------------------------------------------------------------------
        ///<summary>
        /// Establishes a connection to the specified service. This can also be
        /// specified in a WCF configuration file. 
        /// See http://msdn.microsoft.com/en-us/library/ms735103.aspx
        ///</summary>
        //---------------------------------------------------------------------
        private static WindPlantLeadParticipantPortType GetService(EndpointAddress address, X509Certificate2 clientCert, X509Certificate2 serverTrust)
        {
            // Binding for proxy
            BasicHttpBinding binding = new BasicHttpBinding();

            // set security mode to Transport
            binding.Security.Mode = BasicHttpSecurityMode.Transport;
            binding.MessageEncoding = WSMessageEncoding.Text;
            // set credential type to require certificates
            binding.Security.Transport.ClientCredentialType =
                HttpClientCredentialType.Certificate;


            // create channel using binding and address
            ChannelFactory<WindPlantLeadParticipantPortType> factory
                = new ChannelFactory<WindPlantLeadParticipantPortType>(binding, address);

            // assign client side certificate
            factory.Credentials.ClientCertificate.Certificate = clientCert;

            // assign server trusted certificates
            factory.Credentials.ServiceCertificate.DefaultCertificate = serverTrust;

            ServicePointManager.ServerCertificateValidationCallback += delegate(object sender1, X509Certificate certificate,
                                                X509Chain chain, SslPolicyErrors sslPolicyErrors) { return true; };

            // create proxy connection to the server
            return factory.CreateChannel();
        }
    }
}
